"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.KafkaNodeConsumer = void 0;
const events_1 = require("events");
const util = __importStar(require("util"));
const server_services_ordering_zookeeper_1 = require("@fluidframework/server-services-ordering-zookeeper");
const kafka = __importStar(require("kafka-node"));
const kafkaTopics_1 = require("./kafkaTopics");
// time before reconnecting after an error occurs
const defaultReconnectDelay = 5000;
/**
 * Kafka consumer using the kafka-node library
 */
class KafkaNodeConsumer {
    constructor(clientOptions, clientId, groupId, topic, zookeeperEndpoint, topicPartitions, topicReplicationFactor, reconnectDelay = defaultReconnectDelay) {
        this.clientOptions = clientOptions;
        this.groupId = groupId;
        this.topic = topic;
        this.zookeeperEndpoint = zookeeperEndpoint;
        this.topicPartitions = topicPartitions;
        this.topicReplicationFactor = topicReplicationFactor;
        this.reconnectDelay = reconnectDelay;
        this.events = new events_1.EventEmitter();
        clientOptions.clientId = clientId;
        // eslint-disable-next-line @typescript-eslint/no-floating-promises
        this.connect();
        if (zookeeperEndpoint) {
            this.zookeeper = new server_services_ordering_zookeeper_1.ZookeeperClient(zookeeperEndpoint);
        }
    }
    isConnected() {
        return this.client ? true : false;
    }
    /**
     * Returns the offset of the latest consumsed message
     */
    getLatestMessageOffset(partitionId) {
        return undefined;
    }
    async commitCheckpoint(partitionId, queuedMessage) {
        // Although tagged as optional, kafka-node requies a value in the metadata field.
        // Also logs are replayed from the last checkponited offset. To avoid reprocessing the last message
        // twice, we checkpoint at offset + 1.
        const commitRequest = [{
                metadata: "m",
                offset: queuedMessage.offset + 1,
                partition: partitionId,
                topic: this.topic,
            }];
        return new Promise((resolve, reject) => {
            this.consumerGroup.sendOffsetCommitRequest(commitRequest, (err, data) => {
                if (err) {
                    reject(err);
                }
                else {
                    resolve();
                }
            });
        });
    }
    async close() {
        await util.promisify(((callback) => this.consumerGroup.close(false, callback)))();
        await util.promisify(((callback) => this.client.close(callback)))();
        if (this.zookeeperEndpoint) {
            this.zookeeper.close();
        }
    }
    on(event, listener) {
        this.events.on(event, listener);
        return this;
    }
    once(event, listener) {
        this.events.once(event, listener);
        return this;
    }
    async pause() {
        this.consumerGroup.pause();
    }
    async resume() {
        this.consumerGroup.resume();
    }
    async connect() {
        this.client = new kafka.KafkaClient(this.clientOptions);
        const groupId = this.groupId;
        try {
            await (0, kafkaTopics_1.ensureTopics)(this.client, [this.topic], this.topicPartitions, this.topicReplicationFactor);
        }
        catch (error) {
            // Close the client if it exists
            if (this.client) {
                this.client.close();
                this.client = undefined;
            }
            this.events.emit("error", error);
            setTimeout(() => {
                // eslint-disable-next-line @typescript-eslint/no-floating-promises
                this.connect();
            }, this.reconnectDelay);
            return;
        }
        this.consumerGroup = new kafka.ConsumerGroup({
            kafkaHost: this.clientOptions.kafkaHost,
            ssl: this.clientOptions.sslOptions,
            sslOptions: this.clientOptions.sslOptions,
            id: this.clientOptions.clientId,
            autoCommit: false,
            fetchMaxBytes: 1024 * 1024,
            fetchMinBytes: 1,
            fromOffset: "latest",
            groupId,
            maxTickMessages: 100000,
        }, [this.topic]);
        this.consumerGroup.on("connect", () => {
            this.events.emit("connected");
        });
        this.consumerGroup.on("rebalancing", () => {
            const payloads = this.consumerGroup.topicPayloads;
            this.events.emit("rebalancing", this.getPartitions(payloads));
        });
        this.consumerGroup.on("rebalanced", async () => {
            const payloads = this.consumerGroup.topicPayloads;
            const partitions = this.getPartitions(payloads);
            let partitionsWithEpoch;
            try {
                partitionsWithEpoch = await this.fetchPartitionEpochs(partitions);
            }
            catch (err) {
                this.events.emit("error", err);
            }
            this.events.emit("rebalanced", partitionsWithEpoch);
        });
        this.consumerGroup.on("message", (message) => {
            this.events.emit("data", message);
        });
        this.consumerGroup.on("error", (error) => {
            this.events.emit("error", error);
        });
        this.consumerGroup.on("offsetOutOfRange", (error) => {
            this.events.emit("error", error);
        });
    }
    getPartitions(rawPartitions) {
        return rawPartitions.map((partition) => ({
            offset: parseInt(partition.offset, 10),
            partition: parseInt(partition.partition, 10),
            topic: partition.topic,
        }));
    }
    async fetchPartitionEpochs(partitions) {
        let epochs;
        if (this.zookeeperEndpoint) {
            const epochsP = new Array();
            for (const partition of partitions) {
                epochsP.push(this.zookeeper.getPartitionLeaderEpoch(this.topic, partition.partition));
            }
            epochs = await Promise.all(epochsP);
        }
        else {
            epochs = new Array(partitions.length).fill(0);
        }
        const partitionsWithEpoch = [];
        for (let i = 0; i < partitions.length; ++i) {
            const partitionWithEpoch = partitions[i];
            partitionWithEpoch.leaderEpoch = epochs[i];
            partitionsWithEpoch.push(partitionWithEpoch);
        }
        return partitionsWithEpoch;
    }
}
exports.KafkaNodeConsumer = KafkaNodeConsumer;
//# sourceMappingURL=kafkaNodeConsumer.js.map